/// \file standalone.c -- VS1053 Stand-Alone MP3 Player Source Code
/// Copyright (C) 2005-2009 VLSI Solution Oy Tampere Finland

// See readme.txt.

/*
  Note:

  DREQ is changed in normal firmware both in SDI interrupts when the
  stream buffer becomes too full, and whenever the stream buffer becomes
  empty enough.

  So, we can use DREQ ourselves as we wish and access to it does not need to
  be protected by interrupt Disable()/Enable(). However, the normal decoding
  will still keep raising DREQ when it sees that there is space for new data.

- VS1053:
  + I2S
  + higher reference
- IN VS1053b:
  + SCLK = GPIDATA8
  + XCS  = GPIDATA9
  + SI   = GPIDATA10
  + XDCS = GPIDATA11
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "standalone.h"

#if 1 /*for extra inits*/
extern s_int16 __y needCheckNextHeader, haveCachedHeader;
extern u_int32 __y cachedHeader;
struct frame {
#ifdef ALLOC_IN_X
    const struct al_table *alloc;
#else
    __y const struct al_table *alloc;
#endif
    int stereo;
    int jsbound;
    int single;
    int II_sblimit;
    int lsf;
    int mpeg25;
    int header_change;
    int lay;
/*    void (*do_layer)(void);*/
    int error_protection;
    int bitrate_index;
    int sFreqIdx;
    int padding;
    int mode;
    int mode_ext;
    int framesize; /* computed framesize */
};
extern struct frame __y fr;
extern __y struct C1 {
  s_int16 __y fsize;
  s_int16 __y fsizeold;
  s_int16 __y ssize;
  u_int16 __y *bsbuf;
  u_int16 __y bsOffset, bsOldOffset;
  u_int32 __y oldhead;
} c1;
#endif

/* See standalone.h for feature #defines. */

//#define USE_DEBUG

#if defined(SAVE_VOLUME) || defined(SAVE_POSITION)
  #define SAVE_STUFF
#endif

extern volatile s_int16 *srcWp, *srcRp;

extern __y u_int16 g_dcthi[2048];
auto void play_frame(void);
void putch(register __a0 int ch);

#if defined(USE_DEBUG) && !defined(UART_BUFFERED)
  #define ENABLE_SCI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_DAC)|(1<<INT_EN_RX))
  #define ENABLE_SCI_SDI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_SDI)|(1<<INT_EN_DAC)|(1<<INT_EN_RX))
#else
  #define ENABLE_SCI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_DAC)|(1<<INT_EN_RX))
  #define ENABLE_SCI_SDI_DAC_RX ((1<<INT_EN_SCI)|(1<<INT_EN_SDI)|(1<<INT_EN_DAC)|(1<<INT_EN_RX))
#endif

#ifndef RECORDER
// #define FIX_WAV_DECODETIME  /* SCI_UI: advance decodetime for short WAV's */
#endif

#ifndef NOCHECKFILETYPE
  #define CHECKFILETYPE CheckFileType1053
#endif/*NOCHECKFILETYPE*/


s_int16 hcShift;

#ifdef RECORDER /* if standalone player + recorder, disable some features */
  //#undef HDATCHECK           /* no timeout               -14 words */
  //#undef CHECKFILETYPE       /* no filename suffix check -27 words */
  //#undef TIMEOUT
#endif/*RECORDER*/


//#define CTRL3_UPDATE_VOL        (1<<15) /*not needed for vs1053*/
#define CTRL3_BY_NAME           (1<<8) /*locate by name*/
#define CTRL3_RECORD_ON         (1<<7)
#define CTRL3_AT_END            (1<<6)
#define CTRL3_NO_NUMFILES       (1<<5)
#define CTRL3_PAUSE_ON          (1<<4)
#define CTRL3_FILE_READY        (1<<3)

#define CTRL3_PLAY_MODE_MASK    (3<<1)
#define CTRL3_PAUSE_AFTER_PLAY  (3<<1)
#define CTRL3_PAUSE_BEFORE_PLAY (2<<1)
#define CTRL3_LOOP_SONG         (1<<1)
#define CTRL3_PLAY_NEXT         (0<<1)

#define CTRL3_RANDOM_PLAY       (1<<0) /* SCI_AICTRL3 */

#ifndef GPIO_IDATA
  #define GPIO_IDATA GPIO_DATA
#endif


#define SPI_xCS   1 /* GPIO */
#define SPI_CLK   8
#define SPI_MISO  4
#define SPI_MISO_SHIFT -3
#define SPI_xCS2  2
#define SPI_MOSI  1 /* DREQ */



#ifdef USE_DEBUG
static char hex[] = "0123456789ABCDEF";
void puthex(u_int16 d) {
    register i;
    for (i=0;i<4;i++) {
	putchar(hex[(d>>12)&15]);
	d <<= 4;
    }
}
void puthex2(u_int16 d) {
    register i;
    putchar(hex[(d>>4)&15]);
    putchar(hex[(d>>0)&15]);
}
#endif


typedef enum {
    spi_pause=0,
    spi_file=1,
    spi_seek=2,
    spi_waitdata=3,
    spi_data=4,
    spi_zeros=5,
    spi_record=6,
    spi_waitima=7
} SPISTATE;

SPISTATE spiState, oldSpiState;
u_int16 spiCnt;
#ifdef TIMEOUT
u_int16 timeOut;
#endif

struct SPIX { /* 0x1816 do not touch! ASM uses the structure */
    s_int32 size;
    u_int32 sector;
    u_int16 fragSize; /* TODO: make this 32-bit and multiply fragment list clusters by sectorsPerCluster. */
    __y struct FRAGMENT *fragment;
    u_int32 targetSector; /*goto 6,7*/
} spix;


#if 1
auto u_int16 MmcCommand(register __b0 s_int16 cmd, register __d u_int32 arg);
#else
auto u_int16 MmcCommand(register __b0 s_int16 cmd, register __d u_int32 arg) {
    register __b1 u_int16 t;
    #if 1
        puts("Cmd(");
	puthex(cmd);
	puthex(arg>>16);
	puthex(arg);
	puts(")");
    #endif
    /* some MMC brands require the extra clock pulses     */
    SpiSendClocks();
    SpiSendClocks();
    USEX(GPIO_ODATA) &= ~SPI_xCS2;
  
    while (SpiSendReceiveMmc(0xff00,8) != 0xff) {
	/* Unexpected busy signal from MMC */
        #ifdef USE_DEBUG
	    putchar('u');
            putchar('\n');
        #endif
    }
    SpiSendReceiveMmc(cmd, 16); /* top half normally 0xff */
    SpiSendReceiveMmc(arg>>16, 16);
    SpiSendReceiveMmc(arg, 16);
    t = SpiSendReceiveMmc(0x9500, 8);    /* Valid CRC for init, then don't care */
  
    cmd = 100;               /* timeout value, try max. 100 times */
    while (cmd && (t & 0x80)) { /* R1 response not detected */
	t = SpiSendReceiveMmc(0xff00, 8);
	cmd--;               /* decrease the timeout value */
    }
    #ifdef USE_DEBUG
        puthex(t);
        putchar('\n');
    #endif
    return t;
}
#endif

#if 1
auto void ReadDiskSector(register __a u_int32 sector);
#else
auto void ReadDiskSector(register __a u_int32 sector) {
    register s_int16 i;
    register u_int16 *buffer = stream_buffer;
  
    #if 0 && defined(USE_DEBUG)
        //puthex(sector>>16);
        puthex(sector);
        puts(" SECTOR");
    #endif

    MmcCommand(MMC_READ_SINGLE_BLOCK|0x40,(sector&0x7fffffffUL)<<hcShift);
    do {
        i = SpiSendReceiveMmc(0xff00,8);
        #if 0 && defined(USE_DEBUG)
          puthex(i);
          puts(" wait");
        #endif
    } while (i == 0xff);

    #if 0
      if (i != 0xfe) {
          SpiSendClocks();/*USEX(GPIO_ODATA) |= SPI_xCS2;*/
          #ifdef USE_DEBUG
            puts("wfd fail");
          #endif
          return;
      }
    #endif
    
    for (i=0; i<512; i+=2) {
        *buffer++ = SpiSendReceiveMmc(0xffff,16);
    }
    SpiSendReceiveMmc(0xffff,16); /* discard crc */

    /* generate some extra SPI clock edges to finish up the command */
    SpiSendClocks();
    SpiSendClocks();
    return;
}
#endif

#if 0
struct Fat {
    byte BS_jmpBoot[3];         /**< 0 x86 Boot Jump Code */
    byte BS_OEMName[8];         /**< 3 Formatter's name, usually "MSWIN4.1" */
      
    word BPB_BytsPerSec;        /**< 11 Bytes per sector (512) */
    byte BPB_SecPerClus;        /**< 13 Sectors per Cluster (1,2,4,8,..,128) */
    word BPB_RsvdSecCnt;        /**< 14 Reserved sectors (1 (32 for FAT32)) */
    byte BPB_NumFATs;           /**< 16 Number of FATs (2) */
    word BPB_RootEntCnt;        /**< 17 FAT12/16 n of root dir entries */
    word BPB_TotSec16;          /**< 19 Old sector count (0 for FAT32) */
    byte BPB_Media;             /**< 21 Media Type (eg 0xF8) */
    word BPB_FATSz16;           /**< 22 Size of one FAT16 in sectors */
    word BPB_SecPerTrk;         /**< 24 Old CHS Sectors Per Track */
    word BPB_NumHeads;          /**< 26 Old CSH Number of Heads */
    u_32 BPB_HiddSec;           /**< 28 n of sectors before this volume */
    u_32 BPB_TotSec32;          /**< 32 New sector count (0 for FAT12/16) */
     
    /** FAT type specific extensions */
    union Extensions{   
  
        /** FAT12/16 specific extensions to Bios Parameter Block*/
        struct Fat16Specific {    
          byte BS_DrvNum;         /**< 36 DOS INT13 Drive Number (0x80=HD)*/
          byte BS_Reserved1;      /**< 37 For WINNT; Format to 0 */
          byte BS_BootSig;        /**< 38 0x29 if next 3 fields are present */
          byte BS_VolID[4];       /**< 39 Volume ID (usually format datetime) */
          byte BS_VolLab[11];     /**< 43 Volume Label */
          byte BS_FilSysType[8];  /**< 54 Decorative name of fs, eg "FAT16   "*/
        } _16;
  
        /** FAT32 specific extensions to Bios Parameter Block*/
        struct Fat32Specific {    
          u_32 BPB_FATSz32;       /**< 36 Size of one FAT32 in sectors */
          word BPB_ExtFlags;      /**< 40 Flags; active FAT number etc */
          word BPB_FSVer;         /**< 42 File System Version (0x0000) */
          u_32 BPB_RootClus;      /**< 44 Start cluster of Root Dir. (2) */
          word BPB_FSInfo;        /**< 48 Start sector of FSINFO in Resvd area */
          word BPB_BkBootSec;     /**< 50 Sector in Resvd area for BkBoot (6) */
          byte BPB_Reserved[12];  /**< 52 Reserved, Always 0. */
          byte BS_DrvNum;         /**< 64 DOS INT13 Drive Number (0x80=HD) */
          byte BS_Reserved1;      /**< 65 For WINNT, Format to 0 */
          byte BS_BootSig;        /**< 66 0x29 if next 3 fields are present */
          byte BS_VolID[4];       /**< 67 Volume ID (usually format datetime) */
          byte BS_VolLab[11];     /**< 71 Volume Label */
          byte BS_FilSysType[8];  /**< 82 Decorative name of fs, eg "FAT32   "*/
        } _32;
      } ext;
    } fat;
#endif

#if 1
auto u_int16 GetByte(register __c0 u_int16 n);
#else
auto u_int16 GetByte(register __c0 u_int16 n) {
    register u_int16 *p = stream_buffer[n>>1];
    if (n & 1) {
	n = *p;
    } else {
	n = *p >> 8;
    }
    return n & 0xff;
}
#endif
#if 1
auto u_int16 GetWord(register __c0 u_int16 n);
#else
auto u_int16 GetWord(register __c0 u_int16 n) {
    return GetByte(n) | (GetByte(n+1)<<8);
}
#endif
#if 1
auto u_int32 GetLong(register __c0 u_int16 n);
#else
auto u_int32 GetLong(register __c0 u_int16 n) {
    return GetWord(n) | ((u_int32)GetWord(n+2)<<16);
}
#endif

extern __y u_int16 fileName[6];
#ifdef FATINFO_IN_Y
__y struct FATINFO fatInfo;
#else
struct FATINFO fatInfo;
#endif

#if 0 //def USE_DEBUG
void PrintDiskSector(void) {
    register int i;
    register int j;
    for (i=0;i<512;i+=16) {
	puthex(i);
	putchar(' ');
	putchar(' ');
	for (j=0;j<16;j++) {
	    puthex2(GetByte(i+j));
	    putchar(' ');
	}
	putchar(' ');
	for (j=0;j<16;j++) {
	    register int t = GetByte(i+j);
	    if ((t & 0x7f) < 0x20)
		putchar('?');
	    else
		putchar(t);
	}
	putchar('\n');
    }
}
#endif

#if 1
auto u_int16 InitFileSystem(void);
#else
auto u_int16 InitFileSystem(void) {
    /* Load MBR -- the first sector on disk */
    ReadDiskSector(0);
  
    if (GetWord(510) != 0xaa55)
	return 0xaa55;

    #if 0 && defined(USE_DEBUG)
        /* Ok, it should be a MBR sector. Let's verify */
        if (stream_buffer[255] != 0x55aa) {
        #ifdef USE_DEBUG
              puts("!MBR");
        #endif
            return -1; /* sector 0 is not MBR. */
        }
    #endif

#if 0 && defined(USE_DEBUG)
    /* This checks that partition 1 is active. Alter code to allow
       other partition configurations. */
    if (GetByte(0x1be) == 0x80) {
    } else {
	puthex(GetByte(0x1be));
	puts(" PF!");
    }
#endif

    /* Now leave MBR and load sector 0 of partition */
    /* check for boot sector instead of partition table */
    if (GetLong(0x36)!= (((long)'1'<<24)|((long)'T'<<16)|((long)'A'<<8)|'F') &&
        GetLong(0x52)!= (((long)'3'<<24)|((long)'T'<<16)|((long)'A'<<8)|'F')
        ) {
        #ifdef USE_DEBUG
          PrintDiskSector();
          puts("");
        #endif
        fatInfo.currentSector = GetLong(0x1c6);
        ReadDiskSector(fatInfo.currentSector);
    }
    #ifdef USE_DEBUG
      PrintDiskSector();
    #endif
    
    {
        u_int16 BPB_BlksPerSec;

        /* Determine FAT Type (16/32) */
        /* This should be done better, but it'll do for now. */
        fatInfo.BPB_RootEntCnt = GetWord(17);
        /* FAT32 does not have separate root entries. */
        /* TODO: should detect FAT32 by BPB_FATSz16==0 */
        fatInfo.IS_FAT_32 = fatInfo.BPB_RootEntCnt ? 0 : 1;
        fatInfo.FilSysType = GetWord(54+3); /* 0x3231 for FAT12 */
        #ifdef USE_DEBUG
          puthex(fatInfo.FilSysType);
          puts("fsys");
        #endif
        #if 0
          if (!IS_FAT_32){
              if (GetByte(54+4)=='2') {/*diskSect.fat.ext._16.BS_FilSysType[4]=='2'*/) {
                  fatInfo.IS_FAT_12 = GetByte(54);
              }
          }
        #endif
        
        #if 0 && defined(USE_DEBUG)
          if (fatInfo.IS_FAT_32) {
              puts("FAT32");
          } else {
              puthex(fatInfo.BPB_RootEntCnt);
              puts(" FAT");
              puthex(GetWord(19));
              puts(" TSc16");
          }
          puthex(GetWord(54+3)); /* 3231 == '12' 3631= '16' 0000=FAT32 */
          puts(" fst");
        #endif
        #if 1 && defined(USE_DEBUG)
          putchar('\"');
          {
              register int i;
              for (i=0; i<11; i++) {
                  if (fatInfo.IS_FAT_32) {
                      putchar(GetByte(71+i));
                  } else {
                      putchar(GetByte(43+i));
                  }
              }
          }
          putchar('\"');
          putchar('\n');
        #endif
        /* OK, let's calculate */
        /* First, let's get rid of the idea that we have byte addresses
           in the file system. Nope, let's only deal in physical disk
           sectors of 512 bytes. First we convert the FAT byter per sector
           value to "512B disk sectors per fat sector" value. */

        BPB_BlksPerSec = GetWord(11) >> 9;

        /* Then we adjust the Sector per Cluster to mean physical disk
           sectors. in 99% of the cases it is already so because bytes
           per sector almost always is 512 in FAT. Maximum cluster size
           is 65536 bytes (128 disk sectors). */

        fatInfo.fatSectorsPerCluster = GetByte(13)/*BPB_SecPerClus*/ * BPB_BlksPerSec;
        #if 0 && defined(USE_DEBUG)
          puthex(fatInfo.fatSectorsPerCluster);
          puts(" FatSPC");
        #endif
        fatInfo.fatStart = fatInfo.currentSector +
            (u_int32)GetWord(14)/*BPB_RsvdSecCnt*/ * BPB_BlksPerSec;
            #if 0 && defined(USE_DEBUG)
          puthex(fatInfo.fatStart);
          puts(" fatS");
            #endif
        fatInfo.rootStart = GetWord(22);/*BPB_FATSz16*/
        if (fatInfo.rootStart == 0) {
            if (!fatInfo.IS_FAT_32) 
                return 0x0b; /* should be FAT32; can not find root directory */
            fatInfo.rootStart = GetLong(36);/*BPB_FATSz32*/
        }
        /* BPB_NumFATs * BPB_BlksPerSec fits in a word (8+7 bits) */
        fatInfo.rootStart = fatInfo.rootStart * (GetByte(16)/*BPB_NumFATs*/ *
                                                 BPB_BlksPerSec)
            + fatInfo.fatStart;
            #if 0 && defined(USE_DEBUG)
          puthex(fatInfo.rootStart);
          puts(" rootS");
            #endif
        fatInfo.dataStart = fatInfo.rootStart + (fatInfo.BPB_RootEntCnt >> 4)
            - (fatInfo.fatSectorsPerCluster << 1); /*first cluster is cluster 2*/

            #if 0 && defined(USE_DEBUG)
          puthex(fatInfo.dataStart);
          puts(" dataS");
  //      puthex(fatInfo.clusters);
  //      puts(" Clustrs");
            #endif
    }
    return 0;
}
#endif

#if 0
/** FAT/VFAT directory record union */
union DirRecordUnion {
    /** Standard directory entry */
    struct Entry {
      byte Name[11]; 0
      byte Attr;     11
      byte NTRes;    12
      byte CrtTimeTenth; 13
      word CrtTime;  14
      word CrtDate;  16
      word LstAccDate; 18
      word FstClusHi; 20
      word WrtTime;   22
      word WrtDate;   24
      word FstClusLo; 26
      u_32 FileSize;  28
    } entry;
  
    /** Extended directory entry */
    struct LongEntry {
      byte Ord;      0
      word Name1[5]; 1 /**< characters 1-5 */
      byte Attr;     11
      byte Type;     12 /**< entry type, zero=long name component */
      byte Chksum;   13
      word Name2[6]; 14 /**< characters 6-11 */
      word FstClusLO; 26 /**< zero for long entry */
      word Name3[2]; 28 /**< characters 12-13 */
    } longentry;
  };
#endif


//#define MAX_FRAGMENTS 28//16//22
__y struct FRAGMENT fragments[MAX_FRAGMENTS];

#if 1
auto __y struct FRAGMENT *FragmentList(register __i2 __y struct FRAGMENT *frag,
				       register __b u_int32 fatCluster);
#else
auto __y struct FRAGMENT *FragmentList(register __i2 __y struct FRAGMENT *frag,
				       register __b u_int32 fatCluster) {
    register __d0 u_int16 offset;
  
    /* fatCluster means now current cluster, starting from 0, not 2. */
    /* note that startcluster=2 is already compensated in fatInitGlobals */
    #if 0 && defined(USE_DEBUG)
        puthex(fatCluster>>16);
        puthex(fatCluster);
        puts(" fatC");
    #endif
    #if 0
        frag->size = 0;
        frag->start = fatCluster * fatInfo.fatSectorsPerCluster + fatInfo.dataStart;
    #endif
    goto l55; /* read first sector */
  
    while (1) {
	u_int32 t;
	if (fatInfo.IS_FAT_32) {
	    t = GetLong(offset) & 0x0fffffffUL;
	} else {
            #if 0 && defined(USE_DEBUG)
	    {
		register int i;
		for (i=0;i<256;i++) {
		    t = GetByte(i);
		    puthex(t);
		    putchar(' ');
		    if ((i&7)==7)
			putchar('\n');
		}
	    }
            #endif

            t = GetWord(offset);
            if (t >= 0xfff8U)
                t = 0x0fffffffUL;
        }
        #if 0 && defined(USE_DEBUG)
          puthex(t>>16);
          puthex(t);
          puts(" c");
        #endif
        frag->size += fatInfo.fatSectorsPerCluster;
        ++fatCluster;
        if (frag->size >= 32768U || t != fatCluster) {
            /* if not sequential cluster, start next fragment */
            #if 0 && defined(USE_DEBUG)
              puts(" frag");
            #endif
            fatCluster = t;
            if (frag == &fragments[MAX_FRAGMENTS-1] ||
                t >= 0x0ffffff8UL) {
                frag->start |= LAST_FRAGMENT;
                return frag+1;
            }
            frag++;
        l55:
            frag->size = 0;
            frag->start =
                fatCluster * fatInfo.fatSectorsPerCluster + fatInfo.dataStart;

        } else {
            offset += (fatInfo.IS_FAT_32 ? 4 : 2);
            offset &= 511;
            if (offset != 0)
                continue;
        }
read:
        /* If FAT page is not already in memory, load it. */
        if (fatInfo.IS_FAT_32) {
            offset = ((u_int16)fatCluster & 0x7F) << 2;
            ReadDiskSector((fatCluster >> 7) + fatInfo.fatStart);
        } else {
            offset = ((u_int16)fatCluster & 0xff) << 1;
            ReadDiskSector((fatCluster >> 8) + fatInfo.fatStart);
        }
    }
}
#endif

u_int16 gFileNum[2];
#ifdef RECORDER
struct REC {
    u_int16 record;
    u_int16 rd;
    u_int16 wr;
    s_int16 fill;
} record;
#endif/*RECORDER*/
extern __y s_int16 mallocAreaY[9216];

#ifdef __VSDSP__
auto s_int16 OpenFile(register __c0 u_int16 fileNum);
#else
auto s_int16 OpenFile(register __c0 u_int16 fileNum) {
    register __i1 __y struct FRAGMENT *curFragment = &fragments[0];
    register __i0 __y struct FRAGMENT * __y nextFragment;
    
    /* Start at the start of root directory. */
    
    if (fatInfo.IS_FAT_32) {
	nextFragment = FragmentList(curFragment, 2);
    } else {
	curFragment->start = fatInfo.rootStart | LAST_FRAGMENT;
	curFragment->size  = (fatInfo.BPB_RootEntCnt >> 4); /* /sectorsPerCluster?*/
	nextFragment = curFragment + 1;
    }
    gFileNum[0] = fileNum;
    gFileNum[1] = 0;
    return HandleDir(curFragment, nextFragment);
}
#endif

#ifdef CHECKFILETYPE
#if 0
#define MKID(a,b,c) ((a)|((b)<<8)|((u_int32)(c)<<16))
auto s_int16 CheckFileType(register __a u_int32 name) {
    name &= 0x00ffffffUL;
    if (name == MKID('W','A','V'))
	return -1;
    if (name == MKID('M','P','3'))
	return -1;
    if (name == MKID('M','I','D'))
	return -1;
    if (name == MKID('W','M','A'))
	return -1;
    if (name == MKID('A','S','F'))
	return -1;
    if (name == MKID('A','A','C'))
	return -1;
    if (name == MKID('M','P','4'))
	return -1;
    if (name == MKID('M','4','A'))
	return -1;
    if (name == MKID('3','G','P'))
	return -1;
    if (name == MKID('3','G','2'))
	return -1;
    if (name == MKID('O','G','G'))
	return -1;
    return 0;
}
#endif/*1*/
#endif/*CHECKFILETYPE*/

/*147 words*/
auto s_int16 HandleDir(register __i2 __y struct FRAGMENT *curFragment,
		       __y struct FRAGMENT *nextFragment) {
    register __c1 int i = 0;
    register u_int32 currentSector = curFragment->start /*& 0x7fffffffUL */;
  
    #ifdef ENABLE_WATCHDOG
        USEX(WDOG_RESET) = WDOG_RESET_VAL;
    #endif
#pragma msg 36 off
    goto first;
  
    while (1) {
	register __d1 u_int16 fn = GetByte(i+0); /* first char of filename */
  
#pragma msg 36 on
	/* We are now looking at FAT directory structure. */
	/* Is current file a regular file? */
	if (fn == 0)
	    goto eod; /*end of directory*/
	if (fn != 0xe5) { /* Not deleted */
            #ifdef WITH_SUBDIRECTORIES
	    /* Is it a subdirectory? */
	    if ((GetByte(i+11/*Attr*/) & 0x10) && fn != '.') {
		/* TODO: do not call if FAT12! */
		if (HandleDir(nextFragment,
			      FragmentList(nextFragment,
					   ((u_int32)GetWord(i+20)<<16)+GetWord(i+26))) < 0)
		    return -1;
		memset(stream_buffer, 0, 512/2);
		ReadDiskSector(currentSector);/*reread sector*/
	    } else /* it was a subdirectory */
            #endif
		/* Attributes: NO directory, NO volume id, NO system, NO hidden */
	    if ((GetByte(i+11/*Attr*/) & 0xde) == 0
            #ifdef CHECKFILETYPE
                    /*
                      check supported files from name suffix:
                      vs1053: wav mp3 mid wma asf mp4 aac m4a ogg fla
                    */
		&& CHECKFILETYPE(GetLong(i+8))
            #endif
		) {
		u_int32 cluster =
		    ((u_int32)GetWord(i+20) << 16) + GetWord(i+26);
#ifdef RECORDER //12 words
		/* remember the highest used cluster */
		if (mmc.usedTopCluster < cluster) {
		    mmc.usedTopCluster = cluster;
		}
#endif
		/* It is a regular file. */
                #if defined(SCI_UI)
		  /* If open-by-name, check name and open file if match. */
#pragma msg 240 off
		  if (USEX(SCI_AICTRL3) & CTRL3_BY_NAME) {
		    register int j, t = 0;
		    register u_int16 *p = stream_buffer + i/2;
		    register __y u_int16 *d = fileName;
		    p[5] &= 0xff00; /* clear the attribute byte */
		    for (j=0; j<6; j++) {
			t |= *d++ ^ *p++;
		    }
		    if (t == 0) {
			USEX(SCI_AICTRL0) = gFileNum[1];
			goto thisfile;
		    }
		  }
#pragma msg 240 on
                #endif

                gFileNum[1]++; /* total file count */
                if (gFileNum[0]-- == 0) {
                thisfile:
                  #ifdef SCI_UI
		    /* Put the 8.3-character filename to memory so the user
		       can read it. Two bytes per word, big-endian. */
		    {   /* 13 words */
			register int j;
			register u_int16 *p = stream_buffer + i/2;
			register __y u_int16 *d = fileName;
			for (j=0;j<6;j++) {
			    *d++ = *p++;
			}
		    }
                  #endif
                    /* ------------ FILE FOUND ------------- */
                    fatInfo.fileSize = GetLong(i+28); /*do first!*/
                    FragmentList(&fragments[0], cluster);
                    return -1; /* File found, All OK return */
                }
            } /* normal file */
        } /* 0xe5 / 0x00 deleted or unused entry */
        i = (i + 32) & 511;
        if (i == 0) {
            currentSector++;
            if (--(curFragment->size) == 0) {
                if ((s_int16)(curFragment->start>>16) < 0) {
		eod:
                #if defined(SCI_UI)
		    USEX(SCI_AICTRL3) &= ~CTRL3_BY_NAME;
                #endif
                    return (s_int16)gFileNum[1];
                }
                curFragment++;
                currentSector = curFragment->start /*& 0x7fffffffUL*/;
            }
        first:
            #if 0 && defined(USE_DEBUG)
                puthex(currentSector);
                puts(" read");
            #endif
	    memset(stream_buffer, 0, 512/2);
            ReadDiskSector(currentSector);
        }
    }
}


#ifdef HDATCHECK
  u_int16 hdatCheck;
#endif
extern __y u_int16 hwSampleRate;

/* Can be done for all formats, because IdleHook is run as the main thread
   and not in interrupts. Whenever IdleHook is called, there either is
   not enough stream data anyway, or the decoder is outputting samples
   and there is not enough space in the audio buffer (thus not reading
   stream). */
u_int16 __x * __x stream_save;

#if defined(ENABLE_UI) || defined(SCI_KEYS)
__y struct {
    u_int16 Data; /* must be first entry! Used by c.s */
    u_int16 Cnt;
    u_int16 Strobe;
    u_int16 Timer; /* Data,Cnt,Strobe,Timer must be in this order (buts.s) */
    u_int16 Key;
    s_int16 KeyTime;
} ui;
#endif

#if 1
  #ifdef RECORDER
    #ifdef ENABLE_UI
      s_int16 ApplAddrRec(register __i0 s_int16 **d, register __a1 s_int16 mode,
			  register __a0 s_int16 n);
    #endif/*ENABLE_UI*/
    #ifdef SCI_UI
      s_int16 ApplAddrSciRec(register __i0 s_int16 **d, register __a1 s_int16 mode,
                               register __a0 s_int16 n);
    #endif/*SCI_UI*/
  #endif/*RECORDER*/
  
  #if defined(ENABLE_UI) || defined(SCI_KEYS)
    s_int16 ApplAddr(register __i0 s_int16 **d, register __a1 s_int16 mode,
                     register __a0 s_int16 n);
    s_int16 ApplAddrButRef(register __i0 s_int16 **d, register __a1 s_int16 mode,
                     register __a0 s_int16 n);
  #endif/*ENABLE_UI*/

  #ifdef SCI_UI
    s_int16 ApplAddrSci(register __i0 s_int16 **d, register __a1 s_int16 mode,
                        register __a0 s_int16 n);
    s_int16 ApplAddrRef(register __i0 s_int16 **d, register __a1 s_int16 mode,
                        register __a0 s_int16 n);
  #endif/*SCI_UI*/
#else/*1*/
  #ifdef ENABLE_UI
    s_int16 ApplAddr(register __i0 s_int16 **d, register __a1 s_int16 mode,
                     register __a0 s_int16 n) {
        /*Don't check mode, all we get is APPL_AUDIO.*/
        //if (mode == APPL_AUDIO)
        {
            register u_int16 t = hwSampleRate/16;
            /* generate a timebase for User Interface */
            ui.Cnt += n;
            if (ui.Cnt > t) {
                ui.Cnt -= t;
                ui.Timer++;
                ui.Strobe = 1;
            }
        }
    }
  #endif/*ENABLE_UI*/
#endif


#if 1
auto u_int32 NextSector(void);
#else
auto u_int32 NextSector(void) {
    spix.sector++;
    if (spix.fragSize-1 == 0) {
	if ((s_int32)spix.sector < 0) /* was last fragment */
	    return -1;
	spix.fragment++;
	spix.sector  = spix.fragment->start /*& 0x7fffffffUL*/;
	spix.fragSize = spix.fragment->size; /* *fatInfo.sectorsPerCluster*/
    } else {
	spix.fragSize = spix.fragSize-1;
    }
    return 0;
}
#endif

#if 0 && defined(RECORDER)
/* Note: sector is not logical sector number, it is directly MMC argument */
auto void WritePhysicalSectorY(register __d u_int32 sector,
			       register __i3 __y u_int16 *dataBufPtr) {
    register u_int16 c;
    s_int16 to = 0;

    c = MmcCommand(MMC_WRITE_BLOCK|0x40, sector & 0x7fffffff);
    while (c != 0x00) {
	//wait for BUSY token, if you get 0x01(idle), it's an ERROR!
	c = SpiSendReceiveMmc(0xff00, 8);
	to++;
	if (c == 0x01 || to < 0) {
	    return; //ERROR
	}
    }
    //MMC_XCS = MMC_SELECTED;
    SpiSendReceiveMmc(0xfffe, 16);
    for (c=0;c<256;c++) {
	SpiSendReceiveMmc(*dataBufPtr++, 16);
    }
    SpiSendReceiveMmc(0xffff, 16); /* send CRC */
    
    //c = SPIGetChar(); //This prints xxx00101, (usually e5) when data ok
    #ifdef USE_DEBUG
      c = SpiSendReceiveMmc(0xff00, 8);
      puthex2(c);
      puts(" response");
    #endif
    #if 0
      while ( SpiSendReceiveMmc(0xff00, 8) != 0xff)
	  ;
      SpiSendReceiveMmc(0xffff, 16);
      SpiSendReceiveMmc(0xffff, 16);
      SpiSendClocks();
      SpiSendClocks();
    #endif
} 
#endif/*0 && RECORDER*/

#if 1
auto u_int16 swap(register __a0 u_int16 val);
#else
auto u_int16 swap(register __a0 u_int16 val) {
    register u_int32 v = val<<8;
    return (u_int16)(v>>16)|(u_int16)v;
}
#endif


#if defined(FIX_WAV_DECODETIME) && defined(SCI_UI)
u_int16 lastDecodeTime;
#endif /*FIX_WAV_DECODETIME + SCI_UI*/

void SaveWord(register __c0 u_int16 addr, register __c1 u_int16 value);
#ifdef SAVE_VOLUME
int volChanged = 0;
#endif

#ifdef SAVE_STUFF
  #define SPI_WREN  0x06
  #define SPI_WRDI  0x04
  #define SPI_RDSR  0x05
  #define SPI_WRSR  0x01
  #define SPI_READ  0x03
  #define SPI_WRITE 0x02
  
  register __a1 short SpiDelay(void);
  auto u_int16 SpiSendReceiveSpi(register __a0 u_int16 datTopAligned, register __a1 s_int16 bits);
  
  #if 1
  void SaveWord(register __c0 u_int16 addr, register __c1 u_int16 value);
  #else
  void SaveWord(register __c0 u_int16 addr, register __c1 u_int16 value) {
      USEX(GPIO_ODATA) &= ~(SPI_xCS2|SPI_xCS);
      SpiSendReceiveSpi(SPI_WREN<<8, 8);
      USEX(GPIO_ODATA) |= SPI_xCS;
      USEX(GPIO_ODATA) &= ~SPI_xCS;
      SpiSendReceiveSpi(SPI_WRITE<<8, 8);
      SpiSendReceiveSpi(addr/*volume=26*/, 16);
      SpiSendReceiveSpi(value, 16);
      USEX(GPIO_ODATA) |= SPI_xCS|SPI_xCS2;//SPI_CLK;
  }
  #endif
  //SaveWord(26/*volume=26*/, USEX(SCI_VOL));
#endif

#ifdef SHUFFLE_PLAY
u_int16 shuffle_offset = 0;
#endif

#ifdef USE_REWIND
int rewmode = 0;
int rewcount = 0;
#endif

// Definitions for Standalone Legacy Keys (VS1011/VS1003/VS1053) SW1, SW2, SW3
// These are read using special diode connections to the spi bus

#define SW1 0xffffU
#define SW2 0x8000U
#define SW3 0x7fffU
//#define SW4 0x4000U
      

/*
  Called before HALT
 */
void UserHook(void) {
    register __c1 SPISTATE cachedSpiState = spiState;

    #ifdef RECORDER
      #ifdef UART_BUFFERED
        if (UartFill() >= 1) {
            /* All commands are 3 bytes long. */
            register int c = UartPeekByte();
            if (c == 'D' /*0x44 bb bb*/) {
                if (UartFill() >= 3) {
                    register int arg;
                    UartGetByte();
                    arg = UartGetByte();
                    arg = (arg<<8) | UartGetByte();
                    /* Set date */
                    created[0] = arg;
                    putch('d');
                }
            } else if (c == 'T' /*54 bb bb*/) {
                if (UartFill() >= 3) {
                    register int arg;
                    UartGetByte();
                    arg = UartGetByte();
                    arg = (arg<<8) | UartGetByte();
                    /* Set time */
                    created[1] = arg;
                    putch('t');
                }
            } else if (c == 'S' /*53 bb bb*/) {
                if (UartFill() >= 3) {
                    register int arg;
                    UartGetByte();
                    arg = UartGetByte();
                    arg = (arg<<8) | UartGetByte();
                    /* Set seconds */
                    created[2] = arg;
                    putch('s');
                }
            } else {
                UartGetByte();
                /* unknown command */
                putch('u');
            }
        }
      #endif
    #endif/*RECORDER*/
    
    
    #ifdef SCI_UI
      /* play pause/continue */
      if ((USEX(SCI_AICTRL3) & CTRL3_PAUSE_ON)) {
          if (cachedSpiState != spi_pause) {
              oldSpiState = cachedSpiState;
              #ifdef UART_UI
                putch('p');//USEX(UART_DATA) = 'p';
              #endif
              cachedSpiState = spi_pause;
  
              stream_save = stream_wr_pointer;
              stream_wr_pointer = stream_rd_pointer+1;
              /* Does not matter if it is one off..
                 StreamDiff() returns the right value regardless. */
          }
      } else {
          if (cachedSpiState == spi_pause) {
          #ifdef UART_UI
                putch('c');//USEX(UART_DATA) = 'c';
          #endif
              cachedSpiState = oldSpiState;
              stream_wr_pointer = stream_save;
          }
      }
      #ifdef RECORDER
        if ((USEX(SCI_AICTRL3) & CTRL3_RECORD_ON)) {
            if (cachedSpiState != spi_waitima && cachedSpiState != spi_record) {
                record.record = 0x1234;
                goto out_of_wav;
            }
        } else {
            if (cachedSpiState == spi_waitima) {
                USEX(SCI_AICTRL0) = 0;
                goto out_of_wav;
            }
        }
      #endif/*RECORDER*/
    
      if ((USEX(SCI_AICTRL0) & 0x8000U)) {
        /* new song */
        USEX(SCI_AICTRL0) &= 0x7fff; /* ack command */
        /*NEW: clear pause and file ready bits automatically */
        USEX(SCI_AICTRL3) &= ~(CTRL3_PAUSE_ON | CTRL3_FILE_READY | CTRL3_AT_END);

        #ifdef RECORDER
      out_of_wav:
          if (cachedSpiState == spi_waitima) {
              /* wait until MMC write finished */
              while ( SpiSendReceiveMmc(-1/*0xffffU*/, 16) != 0xffffU)
                  ;
  
              SpiSendClocks();
              Disable(); //prevent modu_int from trashing the FAT
              /* Well, it would not because in VS1053 the stream buffer
                 is not used in encode mode. */
  
              mmc.fileSize = spix.size;
              #if 1 /*Fix WAV size fields */
                {
                    register u_int32 t = mmc.fileSize - 8;
                    ReadDiskSector(mmc.freeStart);
                    stream_buffer[2] = SwapWord(t);
                    stream_buffer[3] = SwapWord(t>>16);
                    t -= 0x24;
                    stream_buffer[20] = SwapWord(t);
                    stream_buffer[21] = SwapWord(t>>16);
                          
                    /* We don't care if it fails or not. */
                    MmcWrite(mmc.freeStart, stream_buffer);
                }
              #endif
	      /* FsFatMakeFile() creates the directory entry. */
              FsFatMakeFile();
              //MyReset();
          } else {
              /* especially needed for MIDI!! */
              //USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV);
              //cachedSpiState = spi_zeros;
              /* Discard any unread data, including CRC, before starting again */
              {
                  register int i;
                  for (i=0; i<256+2; i++)
                      SpiSendReceiveMmc(-1/*0xffff*/,16);
              }
	      SpiSendClocks(); //raise chip select
#if 1
	      /*wdog is required to prevent DECIM-restart-bug (channel swap) */
	      clockX = 2;
	      SetRate(10);
	      USEX(WDOG_CONFIG) = 1;
	      USEX(WDOG_RESET) = WDOG_RESET_VAL;
	      while (1)
		  ;
#endif
          }
          #ifdef UART_UI
            putch('e');//USEX(UART_DATA) = 'e';
          #endif
          /* Jump to new track immediately */
          MyReset();
        #else/*RECORDER*/
        
          /* Discard any unread data, including CRC, before starting again */
          {
              register int i;
              for (i=0; i<256+2; i++)
                  SpiSendReceiveMmc(-1/*0xffff*/,16);
          }
	  SpiSendClocks(); //raise chip select
          /* Jump to new track immediately */
          MyReset();
        #endif/*elseRECORDER*/
      }

      /* SCI-controlled version, but with additional key support.
	 Keys are:
	 GPIO4 for Pause/Play(short press) and Toggle Shuffle(long press), 
	 GPIO5 for V+(short) and Next(long), 
	 GPIO6 for V-(short) and Previous(long).

	 20111123
	 1. Keys configuration should be changed to pull low when pressed,
	    and pull high when not pressed.
	    [A: ok. Polarity changed.]

	 2. Do not auto-play song after loading code. Please modify it to
	    receive key input then start to play.
	    [A: Set CTRL3_PAUSE_ON in AICTRL3 before starting code, or
	        read GPIO pins through WRAMADDR/WRAM before starting.)

	 3. Default setting is "always repeat all"
	    [A: this is default.]

	 4. Add shuffle function on Play/Pause key: Play/pause key(short
	    press) and Shuffle enable/disable(long press)
	    [A: Added. Long press toggles shuffle play. When shuffle
	        is activated, a new song is immediately selected.]
       */
      #ifdef SCI_KEYS
        if (ui.Strobe) {
	    ui.Strobe = 0;
#if 0
	    /* pulled low when idle, pulled high when pressed */
	    ui.Data = USEX(GPIO_IDATA) & (7<<4);
#else
	    /* pulled high when idle, pulled low when pressed */
	    ui.Data = ~USEX(GPIO_IDATA) & (7<<4);
#endif
//putch(ui.Data);

	    if (ui.Data == ui.Key) { /* current state same as previous */
		if (ui.KeyTime < 12) {
		    ui.KeyTime++;
		} else if (ui.Key) {
//putch(ui.KeyTime);
		    if (ui.Key == (1<<5)) {
			/* next file */
			USEX(SCI_AICTRL0) = (USEX(SCI_AICTRL0) + 1) | 0x8000U;
//putch('>');
		    } else if (ui.Key == (1<<6)) {
			/* Previous file */
			if (USEX(SCI_DECODE_TIME) < 5) {
			    if (USEX(SCI_AICTRL0) == 0) {
				/* if at first file, go to last file */
				USEX(SCI_AICTRL0) = (USEX(SCI_AICTRL1) - 1)
				    | 0x8000U;
//putch('');
			    } else {
				USEX(SCI_AICTRL0) = (USEX(SCI_AICTRL0) - 1)
				    | 0x8000U; /* previous file */
//putch('<');
			    }
			} else {
			    /* if played 5 seconds or more,
			       start same file from the beginning */
			    USEX(SCI_AICTRL0) |= 0x8000U;
			}
		    } else if (ui.Key == (1<<4)) {
			/* toggle random play */
			USEX(SCI_AICTRL3) ^= CTRL3_RANDOM_PLAY;
			if (USEX(SCI_AICTRL3) & CTRL3_RANDOM_PLAY) {
			    /* when shuffle activated, selects new song
			       immediately. */
			    USEX(SCI_AICTRL0) = 0x8000U;
			}
		    }
		    ui.KeyTime = -32767; /*all long-presses are one-shot*/
		}
	    } else {
		if (ui.Data == 0) {
		    if (ui.KeyTime > 0 && ui.KeyTime < 12) {
			/* was pressed < 0.75 seconds */
			if (ui.Key == (1<<4)) { /*pause/play*/
			    USEX(SCI_AICTRL3) ^= CTRL3_PAUSE_ON;
			} else if (ui.Key == (1<<5)) {
			    if (USEX(SCI_VOL) > 0x0202U) /*volume up*/
				USEX(SCI_VOL) -= 0x0202;
			} else if (ui.Key == (1<<6)) {
			    if (USEX(SCI_VOL) < 0xc0c0U) /*volume down*/
				USEX(SCI_VOL) += 0x0202;
			}
		    }
		}
		ui.Key = ui.Data;
		ui.KeyTime = 0;
	    }
	}
      #endif/*SCI_KEYS*/

    #endif/*SCI_UI*/
    
    #ifdef ENABLE_UI
        
      // PKP: -- UI: USER INTERFACE --
         
      if (ui.Strobe) { // PKP: It's time to check the buttons and do UI calculations
	  /*
	    + SCLK = GPIDATA8
	    + XCS  = GPIDATA9
	    + SI   = GPIDATA10
	    + XDCS = GPIDATA11
	    The following code is compatible with the prototyping board buttons.
	    Code is currently 1 word longer.
	    Connecting the buttons differently 4 keys could be read directly.
	  */
	  ui.Data = ui.Strobe = 0;
                  
	  // PKP: -- UI: READ KEYS --
            
          #ifdef COMPAT_KEYS // PKP: Read the legacy 3-button keys
	    if (!(USEX(GPIO_IDATA) & (1<<10)/*SI*/)) {
		ui.Data = SW2;/*prev/vol-*/
	    } else if (!(USEX(GPIO_IDATA) & (1<<11)/*XDCS*/)) {
		ui.Data = SW1;/*next/vol+*/
	    } else {
		register u_int16 t = USEX(GPIO_ODATA);
		USEX(GPIO_ODATA) = t & ~SPI_xCS;
		/* Should have a short wait here, but seems to work this way.*/
		if (!(USEX(GPIO_IDATA) & (1<<11)/*XDCS*/)) {
		    ui.Data = SW3;/*play/pause/ etc*/
		} else if (!(USEX(GPIO_IDATA) & (1<<9)/*XCS*/)) {
		    ui.Data = SW3;
		}
		USEX(GPIO_ODATA) = t;
	    }
          #else // PKP: Read directly connected keys
	    //USEX(UART_DATA) = USEX(GPIO_IDATA) >> 4;
//#define GERRY
#ifdef GERRY
	    if (!(USEX(GPIO_IDATA) & (1<<10)/*SI*/)) {
		ui.Data = SW3;/*pause/play*/

		if (cachedSpiState == spi_pause) {
		    cachedSpiState = oldSpiState;
				/* restore 'stolen' data */
		    stream_wr_pointer = stream_save;
		    ui.Data = 0;
		}

	    } else if (!(USEX(GPIO_IDATA) & (1<<11)/*XDCS*/)) {
		ui.Data = SW1;/*prev*/
	    } else if (!(USEX(GPIO_IDATA) & (1<<9)/*XCS*/)) {
		ui.Data = SW2;/*next*/
//	    } else if (!(USEX(GPIO_IDATA) & (1<<8)/*SCLK*/)) {
//		ui.Data = SW1;/*prev*/
	    }
#else/*GERRY*/
	    if (!(USEX(GPIO_IDATA) & (1<<10)/*SI*/)) {
		ui.Data = SW2;/*next*/
	    } else if (!(USEX(GPIO_IDATA) & (1<<11)/*XDCS*/)) {
		ui.Data = SW1;/*prev*/
	    } else if (!(USEX(GPIO_IDATA) & (1<<9)/*XCS*/)) {
		ui.Data = SW3;/*pause/play*/
	    } else if (!(USEX(GPIO_IDATA) & (1<<8)/*SCLK*/)) {
		ui.Data = SW1;/*prev*/
//	    } else if ((USEX(GPIO_IDATA) & (1<<4)/*GPIO4*/)) {
//		ui.Data = SW2;
	    }
#endif
          #endif
      
	  // PKP: -- UI: KEY FUNCTIONS --
      
          #ifdef RECORDER/*recorder with buttons*/
	    if (ui.Data == ui.Key) { //PKP: current keystate is same as previous keystate
		if (ui.KeyTime < 12) {
		    ui.KeyTime++;
		} else if (ui.Key) {
		    register u_int16 vol = USEX(SCI_VOL);
		    if (ui.Key == SW1) {
			if (vol) {
			    vol -= 0x0101;
			    goto setvol;
			}
		    } else if (ui.Key == SW2) {
			if (vol != 0xffffU)
			    vol += 0x0101;
		    setvol:
			USEX(SCI_VOL) = vol;
                        #ifdef SAVE_VOLUME
			  volChanged = 1;
                        #endif/*SAVE_VOLUME*/
		    } else if (ui.Key == SW3) {
			if (ui.KeyTime == 12) {
			    if (cachedSpiState != spi_waitima) {
				/* SW3 long press,
				   goto record if not already */
				record.record = 0x1234;

				/* Discard any unread data, including CRC,
				   before starting again */
				{
				    register int i;
				    for (i=0; i<256+2; i++)
					SpiSendReceiveMmc(-1/*0xffff*/,16);
				}
				SpiSendClocks();
				/*
				  Full reset is still required
				  either before or after encoding!
				  We now have it at the end or encoding
				  to reset all peripherals properly.
				*/
				MyReset();
			    }
                            /* If in record mode, end record mode. */
			    ui.KeyTime = -32767; /*one-shot*/
			    goto out_of_wav;
			}
		    }
		}
	    } else {
		if (ui.Data == 0) {
		    if (ui.KeyTime > 0 && ui.KeyTime < 12) {
			/* was pressed < 0.75 seconds */
			if (ui.Key == SW3) {
                                /* play pause/continue */
			    if (cachedSpiState == spi_waitima) {
				/* recording on, stop recording! */
				goto out_of_wav;
			    } else {
				if (cachedSpiState != spi_pause) {
				    oldSpiState = cachedSpiState;
				    cachedSpiState = spi_pause;
        
				    stream_save = stream_wr_pointer;
				    stream_wr_pointer = stream_rd_pointer+1;
				    /* Does not matter if it is one off..
				       StreamDiff() returns the right value
				       regardless. */
				} else {
				    cachedSpiState = oldSpiState;
				    stream_wr_pointer = stream_save;
				}
			    }
			}
			if (ui.Key == SW1 || ui.Key == SW2) {
			    if (ui.Key == SW1) {
				USEX(SCI_AICTRL0) += 1; /* next file */
			    } else {
				/* Previous file */
				if (USEX(SCI_DECODE_TIME) < 5) {
				    if (USEX(SCI_AICTRL0) == 0) {
					USEX(SCI_AICTRL0) = USEX(SCI_AICTRL1) - 1;
				    } else {
					USEX(SCI_AICTRL0) -= 1; /* previous file */
				    }
				}
			    }
                            #if defined(ENABLE_LAYER12)
			      /* gain=0 for partial last frame (bugs) */
			      volume[0] = volume[1] = 0;
                            #endif
			out_of_wav:
			      if (cachedSpiState == spi_waitima) {
				  USEX(SER_DREQ) = 0;
				  /* wait until MMC write finished */
				  while ( SpiSendReceiveMmc(0xffffU, 16) != 0xffffU)
				      ;
				  SpiSendClocks();
				  Disable(); //prevent modu_int from trashing the FAT
				  /* Well, it would not because in VS1053 the stream
				     buffer is not used in encode mode. */
				  
				  mmc.fileSize = spix.size;
				  
                                  /*Fix WAV size fields */
				  {
				      register u_int32 t = mmc.fileSize - 8;
				      ReadDiskSector(mmc.freeStart);
				      stream_buffer[2] = SwapWord(t);
				      stream_buffer[3] = SwapWord(t>>16);
				      t -= 0x24;
				      stream_buffer[20] = SwapWord(t);
				      stream_buffer[21] = SwapWord(t>>16);
                                
				      /* We don't care if it fails or not. */
				      MmcWrite(mmc.freeStart, stream_buffer);
				  }
				  /* FsFatMakeFile() creates the directory
				     entry. */
				  FsFatMakeFile();
                                  #if 1
				    USEX(SCI_AICTRL0) = USEX(SCI_AICTRL3) = 0;
                                  #endif
#if 0
				  /* With vs1053 full reset is not required
				     before or after encoding for it to work
				     more than once, because the decimator
				     (ADC) is not stopped at the end of
				     encoding (and we do not stop it
				     nicely anyway!).
				  */
				  clockX = 2; /*1.0x clock required for wdog*/
				  SetRate(10);
				  USEX(WDOG_CONFIG) = 1;
				  USEX(WDOG_RESET) = WDOG_RESET_VAL;
				  while (1)
				      ;
#else
				  /* so, just use 'soft'reset for vs1053*/
				  MyReset();
#endif
			      }
        
                              /* especially needed for MIDI!! */
			      USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV);
			      cachedSpiState = spi_zeros;

			      /* Discard any unread data, including CRC,
				 before starting again */
			      {
				  register int i;
				  for (i=0; i<256+2; i++)
				      SpiSendReceiveMmc(-1/*0xffff*/,16);
			      }
                              SpiSendClocks(); //raise chip select
			}
		    }
		}
		ui.Key = ui.Data;
		ui.KeyTime = 0;
	    }
	    //ui.Data = 0;
        
        
	    // PKP: Keys
        
	    // SW1 Short Press: Next Song
	    // SW2 Short Press: Previous Song
	    // SW3 Short Press: Pause/Play
        
	    // SW1 Long Press: Volume Up
	    // SW2 Long Press: Volume Down
	    // SW3 Long Press when Playing: Toggle Loudness
	    // SW3 Long Press when Paused: Toggle Shuffle Play
        
        
          #else /* start of non-RECORDER */
	    if (ui.Data == ui.Key) { // PKP: Current Keystate is same as previous Keystate
		if (ui.KeyTime < 8) {
		    ui.KeyTime++;
		} else // PKP: keystate has been consistent for 8 ticks - it's a long keypress
		if (ui.Key
		    && ui.KeyTime > 0
		    ) {
		    /* long presses */
		    register u_int16 vol = USEX(SCI_VOL);
		    if (ui.Key == SW1) {
#ifdef USE_REWIND
			if (rewmode == 0)
			    rewmode = 1;
#else
			if (vol) {
			    vol -= 0x0101;
			    goto setvol;
			}
#endif
		    } else if (ui.Key == SW2) {
			if (vol != (s_int16)0xffffU)
			    vol += 0x0101;
		    setvol:
			USEX(SCI_VOL) = vol; /*in VS1011E disables ints! */
                        #ifdef SAVE_VOLUME
			  volChanged = 1;
                        #endif/*SAVE_VOLUME*/
		    } else {
#ifndef GERRY /* SW3 long press */
			if (ui.KeyTime == 8) {
			    ui.KeyTime = 9;
                            #if defined(LOUDNESS_ALWAYS_ON_OFF) && !defined(NO_RANDOM)
			      /* if pause on -- toggle random play */
			      USEX(SCI_AICTRL3) ^= CTRL3_RANDOM_PLAY;
			      if (cachedSpiState == spi_pause) {
				  cachedSpiState = oldSpiState;
				  stream_wr_pointer = stream_save;
			      }
                            #else /*LOUDNESS_ALWAYS_ON_OFF*/
                                
                               #ifdef NO_RANDOM
			         /* toggle loudness - have LSb set when ON */
			         USEX(SCI_BASS) ^= USEX(SCI_AICTRL2);
                               #else/*NO_RANDOM*/
				 if (cachedSpiState
                                   #ifdef SW3_RANDOM_PLAY
				     !=
                                   #else
                                     ==
                                   #endif
				     spi_pause) {
				     /* if pause on -- toggle random play */
				     USEX(SCI_AICTRL3) ^= CTRL3_RANDOM_PLAY;
				 } else {
				     /* toggle loudness - have LSb set when ON */
				     USEX(SCI_BASS) ^= USEX(SCI_AICTRL2);
				 }
                                #endif/*else NO_RANDOM*/
                              #endif/*else LOUDNESS_ALWAYS_ON_OFF*/
			}
#endif
		    }
		}
	    } else { // PKP: Keystate is not the same as previous keystate
		// PKP: Handle short keypresses
		if (ui.Data == 0
                    && ui.KeyTime > 0
		    ) {
		    // PKP: Handle end of short keypress (current keystate ui.Data=0)
		    /* short presses */
		    if (ui.KeyTime < 8) {
			/* was pressed < 0.5 seconds */
        
			if (ui.Key == SW3) {
#ifndef GERRY
			    /* play pause/continue -- not very good for midi.. */
			    if (cachedSpiState != spi_pause) {
                              #ifndef UI_PAUSE_BEFORE_PLAY
				oldSpiState = cachedSpiState;
				cachedSpiState = spi_pause;
				stream_save = stream_wr_pointer;
				stream_wr_pointer = stream_rd_pointer+1;
                              #endif/*!UI_PAUSE_BEFORE_PLAY*/
			    } else {
				cachedSpiState = oldSpiState;
				/* restore 'stolen' data */
				stream_wr_pointer = stream_save;
			    }
#endif
			} else if (ui.Key == SW1 || ui.Key == SW2) {
			    register u_int16 song = USEX(SCI_AICTRL0);
			    if (ui.Key == SW1) {
				song++; /* next file */
			    } else
                            #ifndef UI_PAUSE_BEFORE_PLAY
			      if (USEX(SCI_DECODE_TIME) < 5)
                            #endif
			      {
				  if (song == 0) {
				      song = USEX(SCI_AICTRL1)-1;
				  } else { /* previous file */
				      song--;
				  }
				  /* else the same file */
			      }
			    USEX(SCI_AICTRL0) = song;
        
                            /* especially needed for MIDI!! */
                            #if !defined(NO_MIDI) && (!defined(ENABLE_LAYER12) || !defined(HDATCHECK))
			      USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV);
                            #endif
			    cachedSpiState = spi_zeros;
                            #if defined(ENABLE_LAYER12)
                               /* gain=0 for partial last frame (bugs) */
                               volume[0] = volume[1] = 0;
                            #endif
			}
		    }//(ui.keyTime was < 8)
		}
		ui.Key = ui.Data;
		ui.KeyTime = 0;
	    }
        
            #if defined(SAVE_VOLUME)
	      /*Save volume when volume button is released. */
	      if (ui.Data == 0 &&
		  volChanged &&
		  cachedSpiState == spi_seek) {
		  volChanged = 0;
		  SaveWord(26/*volume=26*/, USEX(SCI_VOL));
	      }
            #endif/*SAVE_VOLUME*/

            //ui.Data = 0;
        #endif /* !RECORDER */
      }
    #endif /*USE_UI*/
    
    #ifdef PLAYTIME
      if (USEX(SCI_DECODE_TIME) > PLAYTIME && cachedSpiState != spi_zeros) {
          USEX(SCI_AICTRL0)++;
          #if !defined(NO_MIDI) && (!defined(ENABLE_LAYER12) || !defined(HDATCHECK))
            USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV);
          #endif
          cachedSpiState = spi_zeros;
      }
    #endif/*PLAYTIME*/
    
    
    #ifdef RECORDER /* 88 words */
      if (cachedSpiState == spi_waitima) {
	  /* spi_waitima: We are waiting for new data to arrive and
	     write data blocks when there is data available and the
	     card is ready for write. */
          if (SpiSendReceiveMmc(-1/*0xffffU*/, 16) == 0xffffU) {
              /* MMC Ready. */
              SpiSendClocks(); // extra clocks with chip select high
  
              /* Note: byteswap is required for linear 16-bit data.
		 This also swaps the header, so the header is originally
		 byte-swapped and this restores the correct order.

		 Because SD cards take longer time to write sectors
		 when they switch from one erase block to another,
		 we need quite a long FIFO for the audio data.
		 The FIFO in the encoding routine is not enough,
		 so we 'receive' data from the encoder's FIFO and
		 store it into our bigger FIFO and save it from there.
	      */
              if (record.fill >= 256) { /* Data in second FIFO */
		  /* Save a sector from the second FIFO */
                  WritePhysicalSectorYSwap(spix.sector<<hcShift,
                                           (__y void *)(mallocAreaY+record.rd));
  
                  spix.sector++;
                  spix.size += 512; /* in bytes! */
                  record.fill -= 256;
                  record.rd += 256;
                  if (record.rd >= sizeof(mallocAreaY))
                      record.rd -= sizeof(mallocAreaY);
              }
	      /* Limit filesize to about 2GB or to the available space. */
              if (spix.size >= 0x7ffffe00 || /*max size about 1 day at 24kHz*/
                  spix.size/512 >= mmc.freeSize ) {
                  /* disk full, end recording. */
                  goto out_of_wav;
              }
          }
  
          /* Fill second FIFO from encoder's FIFO. */
#define READ_SIZE 32 /*in words -- must be >= 22 words (size of WAV header used) and a power of 2 */
          {
              register s_int16 imaFil = (bssWp - bssRp) & (IMA_ENCODE_BSSPACE-1);
              if (imaFil >= READ_SIZE) {
                  s_int16 ne = record.wr + READ_SIZE;
                  if (ne >= sizeof(mallocAreaY))
                      ne -= sizeof(mallocAreaY);
		  if (ne != record.rd) {
                        memcpyYY(mallocAreaY+record.wr, bssRp, READ_SIZE);
                        bssRp += READ_SIZE;
                        if (bssRp >= g_dcthi+IMA_ENCODE_BSSPACE) {
                            bssRp -= IMA_ENCODE_BSSPACE;
                        }
                        record.wr = ne;
                        record.fill += READ_SIZE;
                  }
                  #ifndef SCI_UI
		  /* because encoder does not use WmaStereoCopy(),
		     applAddr does not get called, so we need to
		     update time base to trigger user interface in
		     recording mode. */
                    ui.Cnt += READ_SIZE; /* Mono linear 16-bit */
                    if (ui.Cnt >= USEX(SCI_AICTRL0)/*RECORD_FS*/ /16) {
                        ui.Cnt -= USEX(SCI_AICTRL0)/*RECORD_FS*/ /16;
                        ui.Timer++;
                        ui.Strobe = 1;
                    }
                  #endif
              }
          }
  
          #if defined(LOWER_LOOPBACK_VOL) /* lower loopback volume */
            volume[0] = LOWER_LOOPBACK_VOL;
            volume[1] = LOWER_LOOPBACK_VOL;
          #endif
      } else if (cachedSpiState == spi_record) {
//#define BYTESWAP(a) (((u_int16)(a)>>8)|((a)<<8))
            /* Data save performs byteswap, so header is byte-swapped
	       originally, and ends up with the correct byte order into
	       file! */
            static __y u_int16 waveHeader[] = {
                /* Start with values for 8000Hz 1-ch 16-bit signed WAV */
                0x4952,0x4646,0x0024,0x0000,0x4157,0x4556,0x6d66,0x2074,
                0x0010,0x0000,0x0001,0x0001,
                RECORD_FS, /* 0x401f*/
                0x0000,
                RECORD_FS*2, /*0x803e*/
                ((RECORD_FS*2)>>16), /*0x0000*/
                0x0002,0x0010,0x6164,0x6174,0xffff,0xffff,0x0000,0x0000,
		0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000,0x0000
                /* 10 zero-samples inserted to align to 32-word boundary */
            };
            /* Currently Mono only */
	    /* Fields are already correct when AICTRL0==RECORD_FS. */
	    waveHeader[12] = USEX(SCI_AICTRL0); /* samplerate */
	    waveHeader[14] = USEX(SCI_AICTRL0)*2; /* byterate */
	    waveHeader[15] = USEX(SCI_AICTRL0) >> 15; /* byterate high word */

            memsetY(mallocAreaY, 0, 512/2);
            memcpyYY(mallocAreaY, waveHeader, sizeof(waveHeader));
            record.fill = record.wr = READ_SIZE;
            spix.size = 0; /* calculate file size, in bytes! */
            /* The WAV header is now pre-filled into our second FIFO.
	       Set the first sector to the beginning of the free space,
	       and enter the data read/write state.
	     */
            spix.sector = mmc.freeStart;
            cachedSpiState = spi_waitima;
      }
    #endif/*RECORDER*/
    
    if (cachedSpiState == spi_file) { /* Play file from the beginning. */
        /* extra byte will be read if the file size is odd */
        spix.size = (fatInfo.fileSize+1)/2; /* convert to words */
        cachedSpiState = spi_seek;
setfrag:
        spix.fragment = &fragments[0];
        spix.sector   = fragments[0].start /*& 0x7fffffffUL*/;
        spix.fragSize = fragments[0].size; /*todo: *fatInfo.sectorsPerCluster*/
    }
    #if defined(FIX_WAV_DECODETIME) && defined(SCI_UI)
      if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) == CTRL3_LOOP_SONG &&
          USEX(SCI_HDAT1) == 0x7665/*(('v'<<8)|'e')*/) {
          if (USEX(SCI_DECODE_TIME) < lastDecodeTime)
              USEX(SCI_DECODE_TIME) = lastDecodeTime;
      }
    #endif /*SCI_UI and FIX_WAV_DECODETIME*/

    if (cachedSpiState == spi_seek) {
#ifdef SCI_UI
	if (spix.targetSector &&
	    (USEX(SCI_STATUS) & (1<<SCIST_DO_NOT_JUMP)) == 0) {
#if 0
	    putch('S');
	    putch('e');
	    putch('e');
	    putch('k');
	    putch(' ');
#endif
	    /* can not jump backwards */
	    if (spix.targetSector >= (spix.fragment->start & 0x7fffffff)) {
		s_int32 skip = spix.targetSector - spix.sector;
		/*jump one fragment at a time if necessary*/
		if (skip >= spix.fragSize)
		    skip = spix.fragSize-1;
		spix.sector += skip;
		spix.fragSize -= skip;
		spix.size -= skip*512/2;

		if (spix.sector == spix.targetSector)
		    spix.targetSector = 0;
	    }
	    needCheckNextHeader = 1;
	}
#endif /*SCI_UI*/

#ifdef USE_REWIND
#define REWIND_STEP_SIZE 64 /* should be bitrate-dependant, and should not rewind midi files */
	if (rewmode == 1) {
	    s_int32 skip = spix.fragment->size - spix.fragSize;
	    if (skip > REWIND_STEP_SIZE)
		skip = REWIND_STEP_SIZE;
	    spix.sector -= skip;
	    spix.fragSize += skip;
	    spix.size += skip*512/2;

	    rewcount = REWIND_STEP_SIZE/4; //how much to send before next jump
	    rewmode = 2;
	} else if (rewmode == 2) {
	    if (--rewcount == 0)
		rewmode = 0;
	}
#endif

    #ifndef SCI_UI
	/* In case we are just trying to find something to play,
	   also read keys once in a while..
	   One problem though: this idle hook may not be called in that case.
	*/
	//if (USEX(SCI_HDAT1) == 0)
	{
	    ui.Cnt += 64;
	    if (ui.Cnt > hwSampleRate/16) {
		ui.Cnt -= hwSampleRate/16;
		ui.Timer++;
		ui.Strobe = 1;
	    }
	}
    #endif

	/* Send the SD/MMC read block command */
	MmcCommand(MMC_READ_SINGLE_BLOCK|0x40,
		   (spix.sector & 0x7fffffffUL) << hcShift);
        cachedSpiState = spi_waitdata;
        spiCnt = 0;
        #ifdef HDATCHECK
          if (++hdatCheck == HDATCHECK && USEX(SCI_HDAT1) == 0) {
              /* if file not detected at HDATCHECK/2 kB, skip the rest */
              goto end_of_file;
          }
        #endif
    }

    /*TODO: handle spi_waitdata and spi_data
      in timer interrupt to increase throughput for FLAC */
    /*
      We have sent the read block command and now check if the data
      is available to be read out.
     */
    if (cachedSpiState == spi_waitdata) {
	register u_int16 res = SpiSendReceiveMmc(0xff00,8);
	if (res == 0xfe) { /* 0xfe is the data start token */
	    cachedSpiState = spi_data; /* go to fetch mode */
	    spiCnt = 0;
            #ifdef TIMEOUT
	        timeOut = 0;
            #endif
	    NextSector();
	} else if (res == 0xff) { /* card is still fetching block */
	    if ((s_int16)++spiCnt < 0) {
		MyReset(); /* timeout, reset immediately */
	    }
	} else {
	    /* Otherwise we have Read Error, and data is not transferred */
	    goto end_of_file;
	}
    } else if (cachedSpiState == spi_data) { /* data transfer state */
	/* Loop until the whole file is read, or stream buffer is full,
	   or the next sector needs to be read. */
	while (spix.size > 0) {
	    register int i, j = 32;
	    register u_int16 *wr;
            #if !defined(RECORDER) && defined(USE_FLAC)
                if (USEX(SCI_HDAT1) == (('f'<<8) | 'L')) {
                    register s_int16 t = stream_wr_pointer - stream_rd_pointer;
                    if (t < 0)
                        t += 0x1800;
                    if (t > 0x1800/*STREAM_BUFFER_SZ*/-40)
                        break; /* stream buffer full, leave loop */
                    wr = (void *)stream_wr_pointer;
                    if (j > spix.size)
                        j = spix.size;
                    for (i=0;i<j;i++) {
                        *wr++ = SpiSendReceiveMmc(-1/*0xffffU*/,16);
                        if (wr >= &stream_buffer[0x1800/*STREAM_BUFFER_SZ*/])
                            wr = stream_buffer;
                    }
                } else
            #endif
	    {
		if (StreamDiff() >= STREAM_BUFFER_SZ-40)
		    break; /* stream buffer full, leave loop */
		wr = (void *)stream_wr_pointer;
		if (j > spix.size)
		    j = spix.size;
		for (i=0;i<j;i++) {
		    *wr++ = SpiSendReceiveMmc(-1/*0xffffU*/,16);
		    if (wr >= &stream_buffer[STREAM_BUFFER_SZ])
			wr = stream_buffer;
		}
	    }
	    stream_wr_pointer = (void *)wr;
	    spix.size -= j;//32;
	    spiCnt += 32;
	    if (spiCnt >= 256) {
		/* 256 words read, get next block in file */
		cachedSpiState = spi_seek;
		SpiSendReceiveMmc(-1/*0xffff*/,16); /* discard crc */
		break;
	    }
            #ifdef TIMEOUT
                timeOut = 0;
            #endif
	    /* watchdog clearing here doesn't work with midi */
	}
  
        #ifdef TIMEOUT
            if ((s_int16)++timeOut == -1) {
                spix.size = 0;
            }
        #endif
    
        #ifdef ENABLE_WATCHDOG
	    USEX(WDOG_RESET) = WDOG_RESET_VAL;
        #endif
        if (spix.size <= 0) {
end_of_file:
#if 0
putch('E');
#endif
            #ifdef SCI_UI
            if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) == CTRL3_LOOP_SONG){
                  /* restart the same file immediately */
                  #if defined(FIX_WAV_DECODETIME) && defined(SCI_UI)
                    lastDecodeTime = USEX(SCI_DECODE_TIME)+1;
                  #endif /*SCI_UI and FIX_WAV_DECODETIME*/
                  cachedSpiState = spi_file;
                  #ifdef UART_UI
                    putch('l');//USEX(UART_DATA) = 'l';
                  #endif
              } else
            #endif/*SCI_UI*/
            {
            #ifdef SCI_UI
		/*NEW: pause after play */
		if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) ==
		    CTRL3_PAUSE_AFTER_PLAY) {
		    USEX(SCI_AICTRL3) |= CTRL3_PAUSE_ON | CTRL3_AT_END;
		} else {
		    USEX(SCI_AICTRL0) += 1; /* next file */
		}
		cachedSpiState = spi_zeros;
#if 0
putch('Z');
putch(USEX(SCI_HDAT1)>>8);
putch(USEX(SCI_HDAT1));
#endif
            #else
		{
		    register int m = (USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK);
		    if (m == CTRL3_LOOP_SONG) {
			/* restart the same file immediately */
			cachedSpiState = spi_file;
		    } else if (m == CTRL3_PAUSE_AFTER_PLAY) {
			oldSpiState = spi_zeros;
			cachedSpiState = spi_pause;
		    } else {
#ifdef LOOP_FILES
			/* keep playing the same file */
#else
			USEX(SCI_AICTRL0) += 1; /* next file */
#endif
			cachedSpiState = spi_zeros;
                        #ifdef USE_FLAC
			    if (USEX(SCI_HDAT1) == (('f'<<8) | 'L'))
				USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV);
                        #endif
		    }
		}
            #endif
            }
        }
        #ifdef NO_WMA /* reject WMA files */
            if (USEX(SCI_HDAT1) == (('W'<<8)|'M')) {
		USEX(SCI_AICTRL0) += 1; /* next file */
		goto goreset;//MyReset();
	    }
        #endif
        #ifdef NO_AAC /* reject AAC files */
	    {
		register u_int16 t = USEX(SCI_HDAT1);
		//0x4154(AT) 0x4144(AD) 0x4d34(M4)
		if (t == 0x4154 || t == 0x4144 || t == 0x4d34) {
		    USEX(SCI_AICTRL0) += 1; /* next file */
		    goto goreset;//MyReset();
		}
	    }
        #endif
    } else if (cachedSpiState == spi_zeros) { /* send zeros after the file */
        /* Workaround for WMA resync after stream ended! */
        parametric_x.resync = 0;

	/*TODO: different for flac */
        if (StreamDiff() < STREAM_BUFFER_SZ-40) {
            register int i, z = parametric_x.endFillByte;
            register u_int16 *wr = (void *)stream_wr_pointer;
            for (i=0;i<32;i++) {
                *wr = z;
                if (wr >= &stream_buffer[STREAM_BUFFER_SZ-1]) {
                    wr = stream_buffer;
                } else {
                    wr++;
                }
                ++spiCnt;
                stream_wr_pointer = (void *)wr;
            }
            /* watchdog clearing here doesn't work with midi */
            #ifdef TIMEOUT
              timeOut = 0;
            #endif
        }
        #ifdef TIMEOUT
            if ((s_int16)++timeOut < 0) {
  //putch(USEX(SCI_HDAT1)>>8);
  //putch(USEX(SCI_HDAT1));
		goto goreset;
	    }
        #endif
        
        if (USEX(SCI_HDAT1) == 0 ||
        #if 1
	    USEX(SCI_HDAT1) == 0x574d ||
        #endif
        #ifdef USE_FLAC
	    USEX(SCI_HDAT1) == (('f'<<8) | 'L') ||
        #endif
            spiCnt >= 256+2*2048) { /* was left to max 256 by spi_data */
            /* 2048 words instead of bytes */
goreset:
            /* Discard any unread data, including CRC, before starting again */
            {
                register int i;
                for (i=0; i<256+2; i++)
                    SpiSendReceiveMmc(-1/*0xffff*/,16);
            }
            SpiSendClocks();

            #ifdef UART_UI
                putch('e');//USEX(UART_DATA) = 'e';
            #endif
            MyReset();
        }
        #ifdef ENABLE_WATCHDOG
            goto wdogack;//USEX(WDOG_RESET) = WDOG_RESET_VAL;
        #endif
    } else {
    #ifdef ENABLE_WATCHDOG
  wdogack:
          USEX(WDOG_RESET) = WDOG_RESET_VAL;
    #endif
    }
    
    #ifdef SCI_UI
        USEX(SER_DREQ) = 0;
    #else
    
      #ifdef NO_DREQ_LED
	/* Save space by not updating DREQ. */
      #else /*NO_DREQ_LED*/
        /* 15 words */
        {
            register int t = 0;
    
            if ((ui.Timer & 0x18) == 0) {
                /* DREQ indicates random play status */
                #ifdef NO_RANDOM
                  t = USEX(SCI_BASS); /* has lowest bit set if enabled*/
                #else
                  t = USEX(SCI_AICTRL3);  /* lowest bit is global random play*/
                  #ifdef LOUDNESS_ALWAYS_ON_OFF
                    /* indicate random play */
                  #else /*LOUDNESS_ALWAYS_ON_OFF*/
                  
                    if (cachedSpiState
                    #ifdef SW3_RANDOM_PLAY
                          ==
                    #else
                          !=
                    #endif
                        spi_pause)
                        t = USEX(SCI_BASS); /* has lowest bit set if enabled*/
                  #endif /*else LOUDNESS_ALWAYS_ON_OFF*/
                #endif /*else NO_RANDOM*/
            }
            USEX(SER_DREQ) = t;
        }
      #endif/*!NO_DREQ_LED*/
    #endif/*!SCI_UI*/
    
    spiState = cachedSpiState;
}



void NewSinTest(void);
auto int PatchFrame(void);
void AacSetRate(register __c1 u_int16 rate);/*Patches some AAC problems*/

/*
  TODO: use a timer interrupt for a watchdog timer.
 */
auto void MyMain(register s_int16 __a1 fullReset) {

  #ifdef UART_BUFFERED
    UartFill();
  #endif

  /* Make sure we have the right upper ROM visible. */
  USEX(VS1053_IROM4) &= ~(IROM4_ENABLE);
  /* clear mp4ASC, seems to be needed after FLAC in vs1063, clear here also.*/
  memset((void *)0x1f00, 0, 64); /*Note: different address for vs1053 and 63*/
  USEX(SCI_HDAT1) = 0; /* clear so FLAC is not indicated. */

  #if defined(SCI_UI) && defined(ENABLE_WATCHDOG)
    USEX(WDOG_CONFIG) = 1000;
    USEX(WDOG_RESET) = WDOG_RESET_VAL;
  #endif

  #if defined(SCI_UI)
    /* no files yet */
    USEX(SCI_AICTRL1) = 0;
    memset(&spix, 0, sizeof(spix));
  #endif

    /* For MMC/SD handling the SCI_NEWMODE must be set, and also
       the relevant GPIO pins configured correctly. */
//#define SD_POWER_ENABLE (1<<6)
#define SD_POWER_ENABLE 0
//#define LINE_IN_ENABLE (1<<6)
    USEX(SCI_MODE) = SciModeOr() | (USEX(SCI_MODE) & ~(1<<SCIMB_OUT_OF_WAV));
    USEX(GPIO_ODATA) = SPI_xCS|SPI_CLK|SPI_xCS2|SD_POWER_ENABLE;
    USEX(GPIO_DDR)   = SPI_xCS|SPI_CLK|SPI_xCS2|SD_POWER_ENABLE
    #if defined(ENABLE_I2S)
	/* Enable I2S pins */
	| 0xf0;
      USEX(I2S_CONTROL) = 0x0c
    #endif
	;

    /* InitHardware() sets default UART speed to 9600bps,
       configures clock/PLL according to SCI_CLOCKF,
       initializes stream and audio buffer pointers,
       sets default samplerate (8000Hz),
       clears HDAT0,HDAT1,DECODE_TIME,AUDATA,
       and sets the cosmetic chip ID in parametric_x.chipID .
    */
    InitHardware(); /* no longer clears memory in vs1053 so preserves consts */

  #ifdef RECORDER
    record.record <<= 1; /* 0x1234 to 0x2468 */
  #endif
    
  #ifndef USE_DEBUG
    /* Enable UART RX interrupt to be able to jump to emulator.
       Allows reflashing without MMC/SD inserted. */
    USEX(INT_ENABLE) = (1<<INT_EN_RX);
  #endif
    InitLayer3(); /* clears mp3 synthesis buffers etc. */
    InitCommon(); /* sets mpeg frame pointers stuff */
    //InitReaders();
    InitDecode(); /* sets synthesis phase variable */
    //InitGeneral();
    //AudioInit();

  #if !defined(ENABLE_LAYER12)
    /* If not powerup reset, re-enable analog immediately. */
    if (!fullReset) {
	USEX(SCI_STATUS) &= ~(1<<SCIST_APDOWN2);
    }
    #if defined(ENABLE_WATCHDOG) && defined(KEEP_DRIVERS_OFF)
      else {
	  /*analog power off and analog driver powerdown */
	  USEX(SCI_STATUS) |= (1<<SCIST_APDOWN1)|(1<<SCIST_APDOWN2);
      }
    #endif
  #endif



#if 0
/*quick loopback test --
  note: check that image is smaller than 8kB if you use 8kB EEPROM! */
    USEX(INT_ENABLE) = ENABLE_SCI_DAC_RX;
    USEX(SCI_DECODE_TIME) = 0;
    USEX(SER_DREQ) = 0;


    USEX(SCI_AICTRL0) = 48000U; /*48000Hz*/
    USEX(SCI_AICTRL1) = 1024; /*no auto gain, gain 1.0x */
    //USEX(SCI_AICTRL3) = 4|0; //mode: linear, stereo (line1/mic + line2)
    //USEX(SCI_AICTRL3) = 4|1; //mode: linear, dual channel (line1/mic + line2)
    //USEX(SCI_AICTRL3) = 4|2; //mode: linear, left channel (line1/mic)
    USEX(SCI_AICTRL3) = 4|3; //mode: linear, right channel (line2)
#if 0
    /*note: line input is the default in VS1053, so turn bit off for MIC */
    USEX(SCI_MODE) &= ~(1<<SCIMB_LINE); /* Use mic input */
#else
    USEX(SCI_MODE) |= (1<<SCIMB_LINE); /* Use line input */
#endif

    ImaEncode(); /* loop there until reset */
#else



  #ifdef RECORDER

    #ifdef ENABLE_UI
      /* Calculates ui.Timer and ui.Strobe,
	 does not need to call UserHook anymore */
      applAddr = ApplAddrRec;
    #endif/*ENABLE_UI*/
    
  #else/*RECORDER*/

    #if defined(SCI_UI) && defined(ENABLE_HIGHREF) && !defined(SCI_KEYS)
      /* Keeps settings 1.65V reference. */
      applAddr = ApplAddrRef;
    #endif

    #if defined(ENABLE_UI) || defined(SCI_KEYS)
      #if defined(ENABLE_HIGHREF)
        /* Calculates ui.Timer and ui.Strobe, keeps settings 1.65V reference.*/
        applAddr = ApplAddrButRef;
      #else
        /* Calculates ui.Timer and ui.Strobe. */
        applAddr = ApplAddr;
      #endif
    #endif/*ENABLE_UI*/

  #endif/*elseRECORDER*/

    /* Allow DREQ and enable interrupts, VERY IMPORTANT! */
    //USEX(INT_ENABLE) = 0;
    USEX(INT_GLOB_ENA) = 0;
    USEX(INT_GLOB_ENA) = 0;
    //USEX(INT_GLOB_ENA) = 0;

    /* wake up MMC! -- If this takes too long, watchdog will bite. */

    #ifdef SAVE_POSITION
      /* Clear the current song number so it will reset between cards. */
      if (USEX(SCI_AICTRL0)) /* 9 words */
          SaveWord(28/*AICTRL0*/, 0); //3 words
    #endif


    /* Initialize MMC/SD card, required for both player and recorder. */

    //while (1)
    {
        register s_int16 i, cmd;

	// Start the MMC/SD card
    tryagain: 

#if defined(SD_POWER_ENABLE) && SD_POWER_ENABLE
	/* If required, power down MMC/SD to make it leave SD mode.
	   However, it should not be powered down at each tryagain,
	   because sometimes multiple passes are required to bring
	   the card into operating mode. */
	// MMC reset using SD_POWER_ENABLE
	USEX(GPIO_DDR) |= SD_POWER_ENABLE;
	USEX(SER_DREQ) = 0;
	{
	  u_int16 gpio0_save = USEX(GPIO_ODATA);
	  USEX(GPIO_ODATA) = 1; //all 0 except eecs
	  for (i=0; i<32000; i++) {
	    USEX(GPIO_ODATA) = 1; //all 0 except eecs
	  }
	  USEX(GPIO_ODATA) = gpio0_save | SD_POWER_ENABLE;
	}
#endif

	/* Send clocks to MMC/SD with chip select high
	   -- is required by most cards. */
	for (i=0; i<512; i++) {
	    SpiSendClocks();
	}

        /* MMC Init, command GO_IDLE_STATE should return 0x01 if all is ok. */
        i = MmcCommand(MMC_GO_IDLE_STATE/*CMD0*/|0x40,0);
        if (i != 1)
            goto tryagain;//continue; /* No valid idle response */

        cmd = MMC_SEND_OP_COND|0x40;
	if (MmcCommand(MMC_SEND_IF_COND/*CMD8*/|0x40, 0x00000122/*2.7-3.6V*/)
	    == 1) {
	    /* MMC answers: 0x05 SD answers: 0x01 */
	    /* Read the whole R7 response! */
	    SpiSendReceiveMmc(-1, 32);
	    cmd = 0x40|41; /* ACMD41 - SD_SEND_OP_COND */
	}

        /* MMC Wake-up call, set to Not Idle (mmc returns 0x00)*/
        //i = 0; /* i is 1 when entered .. but does not make the code shorter*/
        while (1) {
            register int c;
	    if (cmd == (0x40|41)) {
		MmcCommand(0x40|55/*CMD55*/,0);
		c = MmcCommand(cmd/*MMC_SEND_OP_COND|0x40*/, 0x40000000UL);
	    } else {
		c = MmcCommand(cmd/*MMC_SEND_OP_COND|0x40*/, 0);
	    }
            if (c == 0)
                break;
            if (++i >= 12000/*< 0*/ || c != 1) {
            #ifdef USE_DEBUG
                  puts("again");
            #endif
                goto tryagain; /* Not able to power up mmc */
            }
        }

	hcShift = 9;
	/* PRETEC's 1G MMC does not seemd to handle CMD58 here! */
	/* Support HC for SD, but not for MMC! */
	if (cmd == (0x40|41) &&
	    MmcCommand(MMC_READ_OCR/*CMD58*/|0x40, 0) == 0) {
	    if (SpiSendReceiveMmc(-1, 16) & (1<<(30-16))) {
		/* OCR[30]:CCS - card capacity select */
		/* HighCapacity, change shift values */
		hcShift = 0;
	    }
	    SpiSendReceiveMmc(-1, 16); /* read the rest of the response */
	}
        
#ifdef RECORDER //12 words
	if (MmcCommand(MMC_SEND_CID/*CMD10*/|0x40, 0) == 0) {
	    register s_int16 *p = stream_buffer;
	    register int t = 3200;
	    while (SpiSendReceiveMmc(0xff00, 8) == 0xff) {
		if (t-- == 0)
		    goto tryagain;
	    }
	    for (i=0; i<8; i++) {
		*p++ = SpiSendReceiveMmc(-1, 16);
//puthex(p[-1]);
		/*
		       Man     Productname   serial#   date
		          App             rev        res    crc7+stop
		  4G:  1D 4144 0000000000 00 0000157A 0 06A E3
		  64M: 02 0000 53444D3036 34 40185439 0 C77 75
		       00 0011 1122223333 44 44555566 6 677 77

		  2G:  02 544D 5341303247 03 9C046901 0 08B 3D Kingston SD
		  1G:  1B 534D 3030303030 10 B1FFC22A 0 08A 59 Extrememory SD
		*/
	    }
	    if (stream_buffer[1] == mmc.cid[1] &&
		stream_buffer[2] == mmc.cid[2] &&
		stream_buffer[3] == mmc.cid[3] &&
		stream_buffer[4] == mmc.cid[4] &&
		stream_buffer[5] == mmc.cid[5] &&
		stream_buffer[6] == mmc.cid[6] &&
		stream_buffer[7] == mmc.cid[7]) {
		/* same card -- use cached information */
	    } else {
		/* different card -- clear all cached values */
		memsetY(&mmc, 0, sizeof(mmc));
		memcpyXY(&mmc.cid, stream_buffer, 8);
#ifdef USE_DEBUG
		puthex(stream_buffer[0]);
		puthex(stream_buffer[1]);
		puthex(stream_buffer[2]);
		puthex(stream_buffer[3]);
		puthex(stream_buffer[4]);
		puthex(stream_buffer[5]);
		puthex(stream_buffer[6]);
		puthex(stream_buffer[7]);
		putch(' ');
		putch('C');
		putch('H');
		putch('G');
		putch('!');
		putch(' ');
#endif
	    }
//puts("=CID");
	}
#endif

        #if 0
          /* Set Block Size of 512 bytes -- default for at least HC */
          /* Needed by MaxNova S043618ATA 2J310700 MV016Q-MMC */
          /* Must check return value! (MicroSD restart!) */
          if (MmcCommand(MMC_SET_BLOCKLEN|0x40, 512) != 0)
              goto tryagain;
          /* All OK return */
        #endif
    }

    // Init File System (FAT)

    /* Detects cards without partition table */
    memset(stream_buffer, 0, 512/2);
    if (InitFileSystem() == 0) {

    #ifdef RECORDER
      #if !defined(SCI_UI)
            ui.Key = 0;
            ui.KeyTime = -32767; /*Ignore record button release*/
      #endif

	/* When recording is selected, record.record is set to 0x1234,
	   and the software is restarted. The initialization shifts
	   record.record, so it will be 0x2468 here if we should go
	   to record mode. */
        if (record.record == 0x2468) {
            /* If record mode, re-enable analog power+drivers immediately. */
            USEX(SCI_STATUS) &= ~((1<<SCIST_APDOWN1)|(1<<SCIST_APDOWN2));
            memset(&record, 0, sizeof(record));
            //record.record = 0;
            //record.rd = record.wr = 0;

//	    USEX(SER_DREQ) = 0;
	    /*
	      FatPrepareForSaving() goes through FAT allocation table and
	      finds the largest contiguous area. It also goes through the
	      root directory and finds a free directory entry and creates
	      a unique filename that is used later when the file entry is
	      created by FsFatMakeFile().
	     */
            FatPrepareForSaving(); /* search free space and a valid name */
            if (mmc.dirLine >= 0) {
                spiState = spi_record;
                #ifdef UART_UI
                  putch('r');//USEX(UART_DATA) = 'r';
                #endif
                
                /* Lower monitor volume during recording,
                   default values will be restored by InitHardware(). */
                /* does not work because StereoCopy not used. */
                //volume[0] = volume[1] = -128;
                /* This one works! */
		//USEX(DAC_VOL) = 0x0808;
                USEX(DAC_VOL) = 0x0606;

                USEX(SCI_AICTRL0) = RECORD_FS;
                USEX(SCI_AICTRL1) = USEX(SCI_WRAM);     /* auto gain */
                USEX(SCI_AICTRL2) = USEX(SCI_WRAMADDR); /* max gain */
                USEX(SCI_AICTRL3) = 4|2; //mode: linear, mono left channel(mic)
                USEX(SCI_MODE) &= ~(1<<SCIMB_LINE); /* Use mic input */
                USEX(SCI_MODE) |= RecordMode(); /* switch to line input? */

                ImaEncode(); /* loop there until reset */
            }
        }
    #endif/*RECORDER*/
        
        
      #ifdef SCI_UI
	if ((USEX(SCI_AICTRL3) & CTRL3_NO_NUMFILES)) {
	  /* Startup is faster if we don't need to find number of files. */
	  USEX(SCI_AICTRL1) = 0x7fff;
	} else
      #endif/*SCI_UI*/
	  {
	    /* determine number of playable files */
	    USEX(SCI_AICTRL1) = OpenFile(0xffffU);
	    //putch(USEX(SCI_AICTRL1)>>8);
	    //putch(USEX(SCI_AICTRL1));
          }
	
      #if !defined(NO_RANDOM) /* if random play feature active */
	if ((USEX(SCI_AICTRL3) & CTRL3_RANDOM_PLAY)) {
            #ifdef SHUFFLE_PLAY /*shuffle 23 words*/
	  USEX(SCI_AICTRL0) = Shuffle(USEX(SCI_AICTRL1), USEX(SCI_AICTRL0));
            #else
                  /* random play (not shuffle!) */
	  USEX(SCI_AICTRL0) = (myrand() % USEX(SCI_AICTRL1));
            #endif
            }
          #endif
        
	/* Open the file to play. The file number is in AICTRL0. */
        if (OpenFile(USEX(SCI_AICTRL0)
        #ifdef SCI_UI
                        &= 0x7fffU
                       //max 32768 files in SCI-controlled version
        #endif
            ) < 0) {
            #if defined(SKIP_ID3V2) /* 48 words */
              /* check for ID3v2 */
              /* An ID3v2 tag can be detected with the following pattern:
                 $49 44 33 yy yy xx zz zz zz zz */
              ReadDiskSector(fragments[0].start);
              if ((GetLong(0) & 0xffffffL) == 0x334449) {
                  /* skip in 16kB resolution, assume tags are smaller than 2MB */
                  /* 0.5kB resolution now */
                  register u_int16 skip = 32 * GetByte(7) + GetByte(8)/4;
                  /* guard against small fragments */
                  if (skip >= fragments[0].size)
                      skip = fragments[0].size-1;
                  fragments[0].start += skip;
                  fragments[0].size  -= skip;
                  fatInfo.fileSize -= (u_int32)skip * 512;
              }
            #endif
/*ADTS is found from FLAC files, and some of them crash the decoder */
	    #if 1 && defined(RECORDER) && defined(USE_FLAC)
	      ReadDiskSector(fragments[0].start);
	      if (GetLong(0) == 0x43614c66) {
		  /* Skip fLaC files in recorder */
		  USEX(SCI_AICTRL0)++;
		  MyReset();
	      }
	    #endif

            #if defined(SCI_UI)
              if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) ==
                  CTRL3_PAUSE_BEFORE_PLAY) {
                  /* Mark file ready and go to pause state in UserHook */
                  USEX(SCI_AICTRL3) |= CTRL3_PAUSE_ON;
              }
              /*NEW: set FILE_READY always */
              USEX(SCI_AICTRL3) = CTRL3_FILE_READY | 
                  (USEX(SCI_AICTRL3) & ~CTRL3_AT_END);
            #endif
            
            #if defined(PAUSE_AT_POWERON)
	      /*pause at power-on, then use pause after play*/
              if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) ==
                  CTRL3_PAUSE_BEFORE_PLAY) {
                  USEX(SCI_AICTRL3) = CTRL3_PAUSE_AFTER_PLAY;
                  oldSpiState = spi_file;
                  spiState = spi_pause;
                  stream_save = stream_wr_pointer;
              } else {
                  spiState = spi_file;
              }
            #else /*PAUSE_AT_POWERON*/
            
              #if defined(UI_PAUSE_BEFORE_PLAY) && defined(ENABLE_UI)
                oldSpiState = spi_file;
                spiState = spi_pause;
                stream_save = stream_wr_pointer;
              #else
		if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) ==
		    CTRL3_PAUSE_BEFORE_PLAY) {
		    oldSpiState = spi_file;
		    spiState = spi_pause;
		    stream_save = stream_wr_pointer;
		} else {
		    spiState = spi_file;
		}
              #endif
            #endif /*else PAUSE_AT_POWERON*/
            
            
            #ifdef HDATCHECK
              hdatCheck = 0;
            #endif
        } else {
        gofirst:
            USEX(SCI_AICTRL0) = 0; /* file not found, jump to first file */
            MyReset();
        }
    } else {
        /* Initialize again */
        goto gofirst;//MyReset();
    }

    #if !defined(RECORDER) && defined(ENABLE_WATCHDOG) && defined(KEEP_DRIVERS_OFF)
      /*otherwise keep it in powerdown until MMC found*/
      USEX(SCI_STATUS) &= ~(1<<SCIST_APDOWN1); /* enable ANALOG power */
    #endif
      
    
    #ifdef SAVE_POSITION
      /* Save current song number to EEPROM */
      SaveWord(28/*AICTRL0*/, USEX(SCI_AICTRL0));
    #endif


#if 0
putch('S');
putch(fatInfo.fileSize>>24);
putch(fatInfo.fileSize>>16);
putch(fatInfo.fileSize>>8);
putch(fatInfo.fileSize);
#endif

    
    USEX(INT_ENABLE) = ENABLE_SCI_DAC_RX;
    USEX(SCI_DECODE_TIME) = 0;

#if 0 //Sine Test
//      USEX(SCI_VOL) = 0;
      USEX(SCI_VOL) = 0x0404;
//      SetRate(44100U);
      SetRate(48000U);
      USEX(SCI_AICTRL0) = 2972; /*1kHz*/
      USEX(SCI_AICTRL1) = 2972;
      CallIROM4(NewSinTest);
#endif
    
    parametric_x.resync = 32767; /* enable resyncs */

    #if defined(FIX_WAV_DECODETIME) && defined(SCI_UI)
      lastDecodeTime = 0;
    #endif /*SCI_UI and FIX_WAV_DECODETIME*/
    
    stream_rd_pointer = stream_wr_pointer = (void*)stream_buffer;
    stream_rd_odd = 0;
    /* Main decode loop -- patched version of ReadFrame() does not fix
       known decoding problems, vs1053b-patches should also be integrated. */
    InitReadFrame();
    while (1) {
	register int j;
        if (
        #if !defined(RECORDER) && defined(USE_FLAC)
	    !(j = PatchFrame())
        #else
	    !(j = ReadFrame())
        #endif
            ) {
#if 0
putch(0);
putch(0);
putch(c1.oldhead>>24);
putch(c1.oldhead>>16);
putch(c1.oldhead>>8);
putch(c1.oldhead>>0);
#endif
            play_frame();
        } else {
#if 0
 putch((u_int16)stream_rd_pointer>>8);
 putch((u_int16)stream_rd_pointer);
 putch(stream_rd_pointer[-2]>>8);
 putch(stream_rd_pointer[-2]);
 putch(stream_rd_pointer[-1]>>8);
 putch(stream_rd_pointer[-1]);
//putch(needCheckNextHeader);
//putch(haveCachedHeader);
#endif
//putch(fr.lay);
//putch(fr.header_change);
#if 0
putch(0);
putch(1);
putch(c1.oldhead>>24);
putch(c1.oldhead>>16);
putch(c1.oldhead>>8);
putch(c1.oldhead>>0);
#endif
            InitReadFrame();
needCheckNextHeader = 0;
        }
	UserHook();
    }
#endif
}
